import os 
import dotenv

# Common
ROOT_PATH = "/".join(os.path.dirname(os.path.abspath(__file__)).split("/")[:-1])
CLI_ENVS = dotenv.dotenv_values(ROOT_PATH + "/env.development")
VERSION = CLI_ENVS["VERSION"]
TRACKER_ADDR = "https://" + CLI_ENVS["SERVER_ADDR"] + "/tracker/"
CLI_NAME = CLI_ENVS["CLI_NAME"]


# Worker
SHARED_PATH = ROOT_PATH + "/share"
WORKER_ENV_PATH = SHARED_PATH + "/worker.env"
if (CLI_NAME == "ain-worker"):
  IMAGE = "ainblockchain/ain-worker:latest"
else:
  IMAGE = "ainblockchain/ain-worker-staging:latest"

WORKER_ENVS = dotenv.dotenv_values(WORKER_ENV_PATH)
HOST_TTYD_PATH = SHARED_PATH + "/ain_worker_ttyd.sock"
WORKER_SERVER_PATH = "http+unix://" + SHARED_PATH.replace("/", '%2F') + "%2Fain_worker.sock"


import subprocess
import dotenv
import requests_unixsocket
import os
import ain.common.constants as constants
from ain.worker.manager.ttyd import Ttyd
from ain.worker.manager.docker import Docker
import ain.common.util as util
import re

# Singleton pattern Class
class Worker():

  Instance = None

  def __init__(self):
    self.session = requests_unixsocket.Session()

  @staticmethod
  def getInstance():
    if Worker.Instance == None:
      Worker.Instance = Worker()
    return Worker.Instance

  def workerNameCheck(self, name):
    if re.match("^[a-zA-Z0-9_-]*$", name) and len(name) >= 5:
      return True
    else:
      return False

  def start(self, name, max_instance_count, price, mnemonic, description, server_ip, gpu):
    
    # check CLI version
    if util.cliVersionCheck() == False:
      print('[!] You have to input "sudo pip3 install --upgrade ' + constants.CLI_NAME + '".')
      return

    if (os.path.exists(constants.HOST_TTYD_PATH)):
      print("[!] You have to input 'sudo ain worker stop.'")
      return
    
    # check options
    if (price < 0): 
      print("[!] price must be greater then zero.")
      return

    if self.workerNameCheck(name) == False:
      print("The worker name ("+ name +") must be at least 5 characters long\n \
             and can be any combination of numbers, alphabets, '-' and '_' ")
      return

    optionsDict = {
      "NAME": name,
      "MAX_INSTANCE_COUNT": str(max_instance_count),
      "PRICE": str(price),
      "MNEMONIC": mnemonic,
      "DESCRIPTION": description,
      "SERVER_IP": server_ip,
      "GPU_BOOL": str(gpu)
    }

    for option in optionsDict:
      if optionsDict[option] == "":
        print('[-] "' + option + '" - empty')
        return
      dotenv.set_key(constants.WORKER_ENV_PATH, option, optionsDict[option])

    # open provider's ttys socket 
    ttydResult = Ttyd.createSocket(constants.HOST_TTYD_PATH)
    # open docker container for ain worker server
    if ttydResult: 
      Docker.createContainer(constants.IMAGE, optionsDict["GPU_BOOL"])

  def info(self): 
    try:
      response = self.session.get(constants.WORKER_SERVER_PATH + "/info")
      return response.json()['id']
    except Exception as e:
      print("[-] worker server error - info")
      return -1
    
  def status(self):
    # Option
    print("[Worker Option]")
    for option in constants.WORKER_ENVS:
      print(option + ": " + constants.WORKER_ENVS[option])
    
    # Worker Instance
    try:
      ids = self.info()
      if (ids == -1): return
      print("[+] Status: Running")
      print("Instance Count: " + str(len(ids)))
      if (len(ids) != 0):
        option = " ".join(ids)
        subprocess.run(["docker", "stats" , option])
    except Exception as e:
      print(e)
      return

  def stop(self):
    # send terminate message to Worker.
    try:
      response = self.session.get(constants.WORKER_SERVER_PATH + "/terminate")
    except Exception as e:
      print("[-] worker server error - terminate")

    # clean socket and Worker docker Container.
    Docker.removeContainer("ain_worker")
    Ttyd.removeSocket(constants.HOST_TTYD_PATH)
    Ttyd.removeSocket(constants.SHARED_PATH + "/ain_worker.sock")

  def log(self):
    basePath = constants.SHARED_PATH + "/log/"
    logFileName = sorted(os.listdir(basePath))[:-1]

    if (len(logFileName) == 0):
      print("[+] does not exist")
      return
    
    targetFile = logFileName[-1]
    
    print("[?] Do you want to see recent log?(y/n)")
    answer = input()
    if (answer == "n"):
      for i in range(len(logFileName)):
        print(str(i) + " - " + logFileName[i])
      print("[?] input number")
      number = int(input())
      targetFile = logFileName[number]
    
    try:
      path = os.path.join(basePath, targetFile)
      if (answer != "n"):
        subprocess.run(["tail", "-f" , path])
      else:
        subprocess.run(["cat", path])
    except Exception as e:
      print("[-] subprocess(log) error")
      print(e)



import os
import ain.common.constants as constants
import docker
import subprocess

class Docker():

  @staticmethod
  def createContainer(image, gpu):
    try:
      print("[+] pull image.")
      client = docker.from_env()
      client.images.pull(image)
      subprocess.call ('sudo docker rmi $(sudo docker images -f "dangling=true" -q)', shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

      if gpu.lower() == "true":
        runtime = "nvidia"
      else:
        runtime = None

      print("[+] create container.")
      client.containers.run(
        image=image,
        command="yarn start",
        detach=True,
        runtime=runtime,
        name="ain_worker",
        volumes={
          '/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'ro'},
          constants.SHARED_PATH:{'bind': "/share", 'mode': "rw"}
        }
      )
        
      print("[+] started ain worker server!")
    except Exception as e:
      print("[-] failed to create container.")
      print(e)
      exit(1)

  @staticmethod
  def removeContainer(name):
    try:
      client = docker.from_env()
      constainer = client.containers.get(name)
      constainer.remove(force=True)
      
    except Exception as e:
      print("failed to remove container. - " + name)
      print(e)


import subprocess
from subprocess import Popen
import os
from time import sleep

class Ttyd():

  @staticmethod
  def createSocket(socketName):
    try:
      Popen(["ttyd" ,"-i" ,socketName ,"/bin/bash"], shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      sleep(1)
      if (os.path.exists(socketName) == False):
        print("[-] failed to create ttyd socket.")
        return False
      
    except Exception as e:
      print("[-] failed to create ttyd socket.")
      print(e)
      return False
      
    print("[+] succeded to create ttyd socket.")
    return True

  @staticmethod
  def removeSocket(socketName):
    try:
      if (os.path.exists(socketName)):
        os.remove(socketName)
        print("[+] succeded to remove ttyd socket. - " + socketName)
    except Exception as e:
      print("[-] failed to remove ttyd socket. - " + socketName)
      print(e)


import click
from ain.worker.worker import Worker
import ain.common.constants as constants

@click.group()
def call():
  pass

@call.command()
@click.argument("command", type=click.Choice(['start', 'stop', 'status', 'log', 'init', 'version', 'restart']))
@click.option("--name", "-n", "name", default=constants.WORKER_ENVS["NAME"], help="Ain Worker Name", type=str)
@click.option("--max-instance-count", "-m", "max_instance_count", default=constants.WORKER_ENVS["MAX_INSTANCE_COUNT"], help="Maximum number of Instances", type=int)
@click.option("--price", "-p", "price", default=constants.WORKER_ENVS["PRICE"], help="ain/h Price", type=float)
@click.option("--mnemonic", "-m", "mnemonic", default=constants.WORKER_ENVS["MNEMONIC"], help="mnemonic", type=str)
@click.option("--description", "-d", "description", default=constants.WORKER_ENVS["DESCRIPTION"], help="Description", type=str)
@click.option("--server-ip", "-s", "server_ip", default=constants.WORKER_ENVS["SERVER_IP"], help="Server IP", type=str)
@click.option("--gpu", "-g", "gpu", default=constants.WORKER_ENVS["GPU_BOOL"], help="Use GPU (true or false)", type=str)
def worker(command, name, max_instance_count, price, mnemonic, description, server_ip, gpu):
  workerInstance = Worker.getInstance()

  if (command == "start"):
    print("[?] Do you want to start? (y/n)")
    if (input().lower() == 'y'):
      workerInstance.start(name, max_instance_count, price, mnemonic, description, server_ip, gpu)
  elif(command == "restart"):
    info = workerInstance.info()
    if (info != -1):
      print("[+] Instance count:" + str(len(info)))
    print("[?] Do you want to restart? (y/n)")
    if (input().lower() == 'y'):
      workerInstance.stop()
      workerInstance.start(name, max_instance_count, price, mnemonic, description, server_ip, gpu)
  elif (command == "stop"):
    info = workerInstance.info()
    if (info != -1):
      print("[+] Instance count:" + str(len(info)))
    print("[?] Do you want to stop? (y/n)")
    if (input().lower() == 'y'):
      workerInstance.stop()
  elif (command == "log"):
    workerInstance.log()
  elif (command == "init"):
    workerInstance.init()
  elif (command == "status"):
    workerInstance.status()
  elif (command == "version"):
    click.echo(constants.VERSION)
    
if __name__ == '__main__':
  call()
